/*
Audio processing module for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#include "options.h"
#include "audio.h"
#include "stream.h"

#include "lamewrap.h"

int audio_drops=0;
int expected_frame_size=0;

static lw_t lw;

static pthread_t thread=0;
static pthread_mutex_t rblock=PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t wakeup=PTHREAD_COND_INITIALIZER;

#define AUDIO_RINGBUF_SIZE 10
static audio_block blocks[AUDIO_RINGBUF_SIZE];
static int rb_pos=0,rb_fill=0;


static inline void audio_fit(audio_block *ab, uint32_t *dest, int dest_count) {
  uint32_t *src=ab->samples;
  int src_count=ab->sample_count;
  int pos,max=src_count*dest_count;

  // no resampling needed if source and dest are equally long
  if(src_count==dest_count) {
    memcpy(dest,src,src_count*4);
    return;
  }

  // a _very_ simple resampler
  for(pos=0; pos<max; pos+=src_count)
    *dest++=src[pos/dest_count];
}


static void* audio_thread(void *arg) {
  int16_t samples[2*SAMPLES_PER_FRAME];
  int64_t timestamp;
  int count,length;
  uint8_t *data;

  while(Running) {
    // wait until there is some work to do
    pthread_mutex_lock(&rblock);
    while(Running && !rb_fill)
      pthread_cond_wait(&wakeup,&rblock);
    pthread_mutex_unlock(&rblock);

    // a wakeup event also occurs upon termination, so react appropriately
    if(!Running) return NULL;

    // resample the current block
    audio_fit(&blocks[rb_pos],(uint32_t*)samples,SAMPLES_PER_FRAME);

    // release the current block
    pthread_mutex_lock(&rblock);
    rb_fill--;
    rb_pos=(rb_pos+1)%AUDIO_RINGBUF_SIZE;
    pthread_mutex_unlock(&rblock);

    // encode audio
    lw_encode_direct(&lw,samples,SAMPLES_PER_FRAME);
    data=lw_get_frames(&lw,&length,&count);
    if(!data) continue;  // no data, no fun
    timestamp=lw.frameno*lw.out_framelength;
    count*=lw.out_framelength;

    // produce OGM output
    // NOTE: this is a little dirty (but well-thought :) hack -- I use negative
    // indexes in lamewrap's buffers ... but this is OK, because in worst case,
    // lw.frame_size is overwritten, which is not needed anyway
    data[-3]=0x88;
    data[-2]=count&0xFF;
    data[-1]=count>>8;
    stream_write(AUDIO_STREAM,&data[-3],length+3,timestamp);
  }
  return NULL;
}


int audio_init(void) {
  lw.in_rate=SAMPLERATE;
  lw.in_channels=2;
  lw.out_rate=aud_samplerate;
  lw.out_channels=aud_channels;
  lw.bitrate=aud_bitrate;
  lw.quality=aud_quality;
  lw_init(&lw);
  expected_frame_size=lw.framesize;
  printf("audio_init(): using LAME %s\n",get_lame_version());
  return 0;
}

int audio_start(void) {
  return pthread_create(&thread,NULL,audio_thread,NULL);
}

int audio_stop(void) {
  void *result;
  pthread_cond_signal(&wakeup);
  return pthread_join(thread,&result);
}

int audio_close(void) {
  lw_done(&lw);
  return 0;
}


audio_block *audio_get_next_block(void) {
  audio_block *ab=NULL;
  pthread_mutex_lock(&rblock);
  if(rb_fill+1>=AUDIO_RINGBUF_SIZE)
    ++audio_drops;
  ab=&blocks[(rb_pos+rb_fill)%AUDIO_RINGBUF_SIZE];
  pthread_mutex_unlock(&rblock);
  return ab;
}

void audio_block_written(void) {
  pthread_mutex_lock(&rblock);
  if(++rb_fill>=AUDIO_RINGBUF_SIZE)
    rb_fill=AUDIO_RINGBUF_SIZE-1;
  else
    pthread_cond_signal(&wakeup);
  pthread_mutex_unlock(&rblock);
}
