/*
Capture module for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#include "options.h"
#include "capture.h"
#include "dvdecode.h"
#include "audio.h"

uint64_t dv_drops=0;
int dv_frameno=0;
int framedrops=0;
int videobuf=0;
int drops_last_second=0;
static int dls_base=0;

static pthread_t thread=0;
static pthread_mutex_t rblock=PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t video_wakeup=PTHREAD_COND_INITIALIZER;

#define DV_RINGBUF_SIZE 10
static dv_frame frames[DV_RINGBUF_SIZE];
static int rb_pos=0,rb_fill=0;


///////////////////////////////////////////////////////////////////////////////
#ifdef CAPTURE_RAW1394 ////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include "raw1394.h"
static raw1394handle_t h1394=0;

static enum raw1394_iso_disposition my_iso_handler(
  raw1394handle_t handle,
  unsigned char *data,
  unsigned int len,
  unsigned char channel,
  unsigned char tag,
  unsigned char sy,
  unsigned int cycle,
  unsigned int dropped
) {
  uint8_t *p;
  int section_type,dif_sequence,dif_block,offset;
  int rb_idx;
  dv_drops+=dropped;

  // ignore packets with payload length != 480 bytes
  if(len!=488) return RAW1394_ISO_OK;

  // extract some header fields
  p=&data[8];
  section_type=p[0]>>5;
  dif_sequence=p[1]>>4;
  dif_block=p[2];

  // beginning of frame? so finalize the last frame
  if(section_type==0 && dif_sequence==0 && dv_frameno++) {
    // extract audio of this frame
    audio_block *ab=audio_get_next_block();
    dvdecode_extract_audio(frames[(rb_pos+rb_fill)%DV_RINGBUF_SIZE].data,ab);
    audio_block_written();

    // advance to next empty frame in the ring buffer (or report a framedrop)
    pthread_mutex_lock(&rblock);
    if(rb_fill+1>=DV_RINGBUF_SIZE)
      ++framedrops;
    else {
      ++rb_fill;
      pthread_cond_signal(&video_wakeup);
    }
    videobuf=rb_fill;

    // set frame number
    rb_idx=(rb_pos+rb_fill)%DV_RINGBUF_SIZE;
    frames[rb_idx].frameno=dv_frameno-1;
    pthread_mutex_unlock(&rblock);

    // (re-)calculate drops-per-second counter
    if(!(dv_frameno%FRAMERATE)) {
      drops_last_second=framedrops-dls_base;
      dls_base=framedrops;
    }
  } else {
    // get ring buffer index to write to
    pthread_mutex_lock(&rblock);
    rb_idx=(rb_pos+rb_fill)%DV_RINGBUF_SIZE;
    pthread_mutex_unlock(&rblock);
  }

  // calculate DIF packet offset in assembled frame
  offset=dif_sequence*12000;
  switch(section_type) {
    case 0: break; // header
    case 1: offset+=(dif_block+1)*80; break;  // subcode
    case 2: offset+=(dif_block+3)*80; break;  // VAUX
    case 3: offset+=(16*dif_block+6)*80; break;  // audio
    case 4: offset+=((dif_block/15)+dif_block+7)*80; break;  // video
    default: return RAW1394_ISO_OK;  // unknown (ignored)
  }
  if(offset>(DV_FRAMESIZE-480)) return RAW1394_ISO_OK;

  // copy DIF data into frame
  memcpy(&frames[rb_idx].data[offset],p,480);
  return RAW1394_ISO_OK;
}


static void* capture_thread(void *arg) {
  if(raw1394_iso_recv_start(h1394,-1,-1,0)<0) {
    printf("capture: cannot start isochronous transfer: %s\n",strerror(errno));
    Running=0;
  }

  while(Running && !raw1394_loop_iterate(h1394));

  raw1394_iso_stop(h1394);
  return NULL;
}


int capture_init(void) {
  h1394=raw1394_new_handle();
  #define PORTINFO_COUNT 16
  struct raw1394_portinfo pinf[PORTINFO_COUNT];
  int port_count,i;

  if(!h1394) {
    printf("capture_init(): cannot get raw1394 handle: %s\n",strerror(errno));
    return 1;
  }

  port_count=raw1394_get_port_info(h1394,pinf,PORTINFO_COUNT);
  if(!port_count) {
    printf("capture_init(): no raw1394 ports found\n");
    raw1394_destroy_handle(h1394);
    return 1;
  }

  for(i=0; i<port_count; ++i)
    printf("capture_init(): port #%d: %d nodes, name = \"%s\"\n",i,pinf[i].nodes,pinf[i].name);

  printf("capture_init(): using port #%d\n",port);
  if(raw1394_set_port(h1394,port)<0) {
    printf("capture_init(): cannot set raw1394 port: %s\n",strerror(errno));
    return 1;
  }

  if(raw1394_iso_recv_init(h1394,my_iso_handler,10000,512,DV_CHANNEL,RAW1394_DMA_DEFAULT,-1)<0) {
    printf("capture_init(): cannot init isochronous transfer: %s\n",strerror(errno));
    return 1;
  }

  return 0;
}

int capture_start(void) {
  return pthread_create(&thread,NULL,capture_thread,NULL);
}

int capture_stop(void) {
  void *result;
  return pthread_join(thread,&result);
}

int capture_close(void) {
  raw1394_iso_shutdown(h1394);
  raw1394_destroy_handle(h1394);
  return 0;
}


#endif ////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
#ifdef CAPTURE_PIPE ///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

static void* capture_thread(void *arg) {
  int rb_idx;
  int rpos;
  
  while(Running) {
    // get frame buffer to write to
    pthread_mutex_lock(&rblock);
    rb_idx=(rb_pos+rb_fill)%DV_RINGBUF_SIZE;
    pthread_mutex_unlock(&rblock);
    frames[rb_idx].frameno=dv_frameno++;
    
    // frame read loop
    rpos=0;
    while(Running && rpos<DV_FRAMESIZE) {
      int count=(int)read(STDIN_FILENO,&frames[rb_idx].data[rpos],DV_FRAMESIZE-rpos);
      if(count>0)
        rpos+=count;
      else if(count<0)
        Running=0;
      else if(benchmark)
        benchmark=2;
    }
    if(!Running) break;

    // extract audio of this frame
    audio_block *ab=audio_get_next_block();
    dvdecode_extract_audio(frames[rb_idx].data,ab);
    audio_block_written();

    // benchmark mode: wait until ring buffer is sufficiently empty
    if(benchmark)
      while(Running && rb_fill) usleep(1000);

    // advance to next empty frame in the ring buffer (or report a framedrop)
    pthread_mutex_lock(&rblock);
    if(rb_fill+1>=DV_RINGBUF_SIZE)
      ++framedrops;
    else {
      ++rb_fill;
      pthread_cond_signal(&video_wakeup);
    }
    videobuf=rb_fill;
    pthread_mutex_unlock(&rblock);

    // (re-)calculate drops-per-second counter
    if(!(dv_frameno%FRAMERATE)) {
      drops_last_second=framedrops-dls_base;
      dls_base=framedrops;
    }    

  }
  return NULL;
}

int capture_init(void) {
  return 0;
}

int capture_start(void) {
  return pthread_create(&thread,NULL,capture_thread,NULL);
}

int capture_stop(void) {
  void *result;
  return pthread_join(thread,&result);
}

int capture_close(void) {
  return 0;
}

#endif ////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

dv_frame *video_get_frame(void) {
  dv_frame *frame=NULL;
  pthread_mutex_lock(&rblock);
  while(Running && !rb_fill)
    pthread_cond_wait(&video_wakeup,&rblock);
  if(Running)
    frame=&frames[rb_pos];
  pthread_mutex_unlock(&rblock);
  return frame;
}

void video_frame_done(void) {
  pthread_mutex_lock(&rblock);
  rb_fill--; videobuf=rb_fill;
  rb_pos=(rb_pos+1)%DV_RINGBUF_SIZE;
  pthread_mutex_unlock(&rblock);
  if(benchmark>1 && !rb_fill) // benchmark mode, EOF
    Running=0;
}

void video_signal_exit(void) {
  pthread_cond_signal(&video_wakeup);
}
