/*
Audio encoder module (LAME wrapper) for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "lame.h"
#include "lamewrap.h"


//////////////////////////////////////////////////////////////////////////////


static inline int get_mp3_framesize(uint8_t *frame) {
  const static int bitrates[2][16]={
    {0,  8, 16, 24, 32, 40, 48, 56, 64, 80, 96,112,128,144,160,0},
    {0, 32, 40, 48, 56, 64, 80, 96,112,128,160,192,224,256,320,0}
  };
  const static int samplerates[2][4]={
    {22050,24000,16000,0x0FFFFFFF},
    {44100,48000,32000,0x0FFFFFFF}
  };
  int version;
  register int b2;
  // reject frames that do not look like MP3
  if ((frame[0]!=0xFF)        // no frame sync byte 10-3
  || ((frame[1]&0xF0)!=0xF0)  // no frame sync bits 2-0, or MPEG 2.5
  || ((frame[1]&0x06)!=0x02)  // not layer 3
  || ((frame[2]&0xF0)==0xF0)  // invalid bitrate
  || (!(frame[2]&0xF0))       // free format
  || ((frame[2]&0x0C)==0x0C)) // invalid sample rate
     return 0;
  // calculate frame size
  version=(frame[1]>>3)&1;
  b2=frame[2];
  return 72000*(version+1)*bitrates[version][b2>>4]/samplerates[version][(b2>>2)&3]+((b2>>1)&1);
}


//////////////////////////////////////////////////////////////////////////////


// NOTE: the resampling functions are somewhat "optimized" for x86's
//       base+index*scale addressing; they may be significantly slower on
//       other architectures


static inline void resample2(int16_t *in, int16_t *out, int count) {
  for(; count; --count, in+=2) {
    register int v=in[0]+in[1];
    *out++=v>>1;
  }
}

static inline void resample4(int16_t *in, int16_t *out, int count) {
  for(; count; --count, in+=4) {
    register int v=in[0]+in[1]+in[2]+in[3];
    *out++=v>>2;
  }
}

static inline void resample8(int16_t *in, int16_t *out, int count) {
  for(; count; --count, in+=8) {
    register int v=in[0]+in[1]+in[2]+in[3]+in[4]+in[5]+in[6]+in[7];
    *out++=v>>3;
  }
}


//////////////////////////////////////////////////////////////////////////////


static void do_encode(lw_t* lw, int16_t *in, int count, int16_t *out) {
  int sscount=0;
  int res;
  if(!in || (lw->subsample && (!out))) return;
  if(lw->subsample) sscount=count*lw->in_channels/lw->subsample;
  switch(lw->subsample) {
    case 2: resample2(in,out,sscount); break;
    case 4: resample4(in,out,sscount); break;
    case 8: resample8(in,out,sscount); break;
    default: // unsupported subsampling mode (or no SS at all) -> encode directly
      res=lame_encode_buffer_interleaved(lw->lame,in,count,&lw->buf[lw->wpos],LW_BUFSIZE-lw->wpos);
      if(res>0) lw->wpos+=res;
      return;
  }
  res=lame_encode_buffer(lw->lame,out,out,sscount,&lw->buf[lw->wpos],LW_BUFSIZE-lw->wpos);
  if(res>0) lw->wpos+=res;
}


static uint8_t *do_get_frames(lw_t* lw, int *length, int *count, int max) {
  uint8_t *retval=NULL;
  if(!lw || lw->rpos>=lw->wpos) return NULL;
  *length=0;
  *count=0;
  
  // if half of the buffer is full, copy the relevant parts to the front
  if(lw->rpos>=LW_BUFSIZE/2) {
    memcpy(lw->buf,&lw->buf[lw->rpos],lw->wpos-lw->rpos);
    lw->wpos-=lw->rpos;
    lw->rpos=0;
  }

  // return consecutive frames  
  while(lw->rpos+4<=lw->wpos) {
    int framesize=get_mp3_framesize(&lw->buf[lw->rpos]);
    if(framesize<4) break;  // no valid frame
    if(lw->rpos+framesize>lw->wpos) break;  // frame is longer than buffered
    if(!retval) retval=&lw->buf[lw->rpos];
    length[0]+=framesize;
    count[0]++;
    lw->frameno++;
    lw->rpos+=framesize;
    if(max && (max>=*count)) break;
  }
  return retval;
}


//////////////////////////////////////////////////////////////////////////////


void lw_init(lw_t* lw) {
  if(!lw) return;

  // create LAME instance
  lw->lame=lame_init();
  if(!lw->lame) return;
  
  // check if we can do the resampling ourselves
  if(((lw->in_rate==lw->out_rate) || (lw->in_rate==lw->out_rate*2) || (lw->in_rate==lw->out_rate*4))
  && ((lw->out_channels==1))) {
    // yes, use internal resampling
    lw->subsample=lw->in_rate*lw->in_channels/lw->out_rate/lw->out_channels;
    lame_set_in_samplerate(lw->lame,lw->out_rate);
    lame_set_num_channels(lw->lame,lw->out_channels);
    lame_set_out_samplerate(lw->lame,lw->out_rate);
    lame_set_mode(lw->lame,3);
  } else {
    // no, let LAME do all the conversion
    lw->subsample=0;
    lame_set_in_samplerate(lw->lame,lw->in_rate);
    lame_set_num_channels(lw->lame,lw->in_channels);
    lame_set_out_samplerate(lw->lame,lw->out_rate);
    if(lw->out_channels==1) lame_set_mode(lw->lame,3);
  }

  // init LAME  
  lame_set_bWriteVbrTag(lw->lame,0);
  lame_set_quality(lw->lame,lw->quality);
  lame_set_brate(lw->lame,lw->bitrate);
  lame_init_params(lw->lame);

  // init LW structures
  lw->in_framelength=(lw->in_rate<32000)?576:1152;
  lw->out_framelength=(lw->out_rate<32000)?576:1152;
  lw->framesize=125*lw->out_framelength*lw->bitrate/lw->out_rate;
  lw->frameno=lw->rpos=lw->wpos=0;
  lw->ssbuf=NULL;
}


void lw_encode_direct(lw_t* lw, int16_t *samples, int count) {
  if(!lw || !lw->lame) return;
  do_encode(lw,samples,count,samples);
}


void lw_encode(lw_t* lw, int16_t *samples, int count) {
  if(!lw || !lw->lame) return;
  if(lw->subsample && (!lw->ssbuf))
    lw->ssbuf=malloc(2*lw->out_rate*lw->out_channels);
  if(count>lw->in_rate) count=lw->in_rate;
  do_encode(lw,samples,count,lw->ssbuf);
}


void lw_done(lw_t* lw) {
  int res;
  if(lw->ssbuf) free(lw->ssbuf);
  if(!lw->lame) return;
  res=lame_encode_flush(lw->lame,&lw->buf[lw->wpos],LW_BUFSIZE-lw->wpos);
  if(res>0) lw->wpos+=res;
  lame_close(lw->lame);
  lw->lame=NULL;
}



uint8_t* lw_get_frame(lw_t* lw, int *length) {
  int count;
  return do_get_frames(lw,length,&count,1);
}

uint8_t* lw_get_frames(lw_t* lw, int *length, int *count) {
  return do_get_frames(lw,length,count,0);
}


///////////////////////////////////////////////////////////////////////////////


#ifdef _TEST_LAMEWRAP
#include <signal.h>

#define SAMPLES 1920
#define BARSIZE 64

int cont;
// a simple C-c handler to abort processing in a safe manner
static void cchandler(int signum) {
  printf("<SIGINT>");
  cont=0; 
}

static void handle_buffer(lw_t *lw, FILE *out) {
  char bar[BARSIZE+1];
  int length,count; uint8_t *data;
  int rb=(lw->rpos*BARSIZE+BARSIZE/2)/LW_BUFSIZE;
  int wb=(lw->wpos*BARSIZE+BARSIZE/2)/LW_BUFSIZE;
  memset(bar,32,BARSIZE); bar[BARSIZE]=0;
  bar[rb]='|'; bar[wb]='|';
  for(count=rb+1; count<wb; ++count) bar[count]='-';
  printf("\n%5d [%s]",lw->frameno,bar);
  data=lw_get_frames(lw,&length,&count);
  if(data && count) printf("%2dx%d",count,length/count);
  fwrite(data,1,length,out);
}

int main(void) {
  FILE *in=fopen("actest.avi.wav","rb");
  FILE *out=fopen("lamewrap_test.mp3","wb");
  lw_t lw;
  fseek(in,SEEK_SET,44);
  
  lw.in_rate=48000;
  lw.in_channels=2;
  lw.out_rate=24000;
  lw.out_channels=1;
  lw.bitrate=40;
  lw.quality=7;
  lw_init(&lw);

  cont=1;  
  signal(SIGINT,cchandler);
  while(cont && (!feof(in))) {
    int16_t smp[SAMPLES*2];
    int count=fread(smp,4,SAMPLES,in);
    lw_encode_direct(&lw,smp,count);
    handle_buffer(&lw,out);
  }
  signal(SIGINT,SIG_DFL);

  lw_done(&lw);
  handle_buffer(&lw,out);
  printf("\nencoded %d frames  %d samples => %.2f seconds\n",lw.frameno,lw.framelength,lw.frameno*lw.framelength/(1.0*lw.in_rate));
  printf("%d bytes remaining in buffer.\n",lw.wpos-lw.rpos);

  fclose(in);
  fclose(out);
  return 0;
}

#endif
