/*
Command line options parser for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#define __INCLUDED_FROM_OPTIONS_C__
#include "options.h"
#include "xvidwrap.h"
#include <ctype.h>

#define DEFAULT_CONFIG_FILE "lvstream.conf"
#define LONGEST_OPTION_NAME "23"

#define streq(a,b) (!strcmp(a,b))

#define OPTN_TYPE_EOL    0
#define OPTN_TYPE_INT    1
#define OPTN_TYPE_TEXT   2
#define OPTN_TYPE_IGNORE 3

typedef struct {
  int optn_type;
  char short_name;
  char* long_name;
  int* optn_ptr;
  char** text_ptr;
  int min,max;
  int default_value;
  char *default_text;
  char *help_text;
} option_t;

#define INT_OPTN(min,max,default,short,name,help) \
        {OPTN_TYPE_INT,short,#name,&name,NULL,min,max,default,NULL,help},
#define TEXT_OPTN(default,short,name,help) \
        {OPTN_TYPE_TEXT,short,#name,NULL,&name,0,0,0,default,help},

static option_t options[]={
  {OPTN_TYPE_IGNORE,'h',"help",NULL,NULL,0,0,0,NULL,     "display this help text"},
  TEXT_OPTN(DEFAULT_CONFIG_FILE,'c',config_file,         "name of config file to use")
#ifdef CAPTURE_RAW1394
  INT_OPTN(0,    15,   0,    'p',port,                   "number of IEEE1394 port to use")
#endif
  INT_OPTN(22050,48000,24000,'R',aud_samplerate,         "resample audio to this rate (Hz)")
  INT_OPTN(1,    2,    1,    'C',aud_channels,           "use mono (1) or stereo (2) audio encoding")
  INT_OPTN(8,    320,  40,   'B',aud_bitrate,            "audio bitrate in kbps")
  INT_OPTN(0,    9,    7,    'Q',aud_quality,            "LAME algorithmic quality 0=best=slowest, 9=worst=fastest")
  INT_OPTN(1,    500,  50,   'k',keyframe_interval,      "keyframe interval in frames")
  INT_OPTN(0,    1,    1,    'f',force_keyframe_interval,"strictly enforce static keyframe interval")
  INT_OPTN(1,    10000,300,  'L',lo_bitrate,             "lo-res video bitrate in kbps")
  INT_OPTN(1,    10000,1500, 'H',hi_bitrate,             "hi-res video bitrate in kbps")
  INT_OPTN(1,    99999,10,   'a',au_length,              "length of an access unit (AU) in seconds")
  INT_OPTN(0,    9,    8,    'q',me_quality,             "motion estimation (ME) quality level; the higher, the better, the slower")
  INT_OPTN(0,    100,  10,   't',framedrop_tolerance,    "tolerated percentage of dropped frames before decreasing ME quality")
  TEXT_OPTN(NULL,            's',sprite_file,            "uncompressed grayscale BMP file to overlay")
  TEXT_OPTN("0,0",           'S',sprite_origin,          "upper-left corner of sprite overlay")
  INT_OPTN(0,    256,  0,    'd',deint_strength,         "deinterlacer strength (recommended: 240-248)")
  TEXT_OPTN(NULL,            'o',output_dir,             "directory to put segmented output in")
  TEXT_OPTN(NULL,            'O',ogm_output,             "file name base to store unsegmented OGM output")
#ifdef CAPTURE_PIPE
  INT_OPTN(0,    1,    0,    'b',benchmark,              "encode as fast as possible and print speed afterwards")
#endif
  {OPTN_TYPE_EOL}
};


static void show_help(void) {
  option_t *o;
  printf("Available command line options:\n");
  for(o=options; o->optn_type; ++o) {
    printf("  \e[1m-%c\e[0m, \e[1m--%-" LONGEST_OPTION_NAME "s\e[0m",o->short_name,o->long_name);
    switch(o->optn_type) {
      case OPTN_TYPE_INT:
        printf("  (integer, range %d-%d, default %d)\n",o->min,o->max,o->default_value);
	break;
      case OPTN_TYPE_TEXT:
        if(o->default_text)
          printf("  (string, default `%s')\n",o->default_text);
        else
          printf("  (string, default: disabled)\n");
        break;
      default:
        printf("\n");
    }
    if(o->help_text)
      printf("      %s\n",o->help_text);
  }
  printf(
"Before parsing the command line, the config file is read. This config file is a\n"
"simple text file with each line containing a long option name without the `--'\n"
"prefix, followed by whitespace and then the option's value.\n"
"The config file overrides the default settings, and the command line options,\n"
"in turn, override the config file settings.\n"
  );
}


void set_default_options(void) {
  option_t *o;
  for(o=options; o->optn_type; ++o)
    if(o->optn_ptr)
      switch(o->optn_type) {
        case OPTN_TYPE_INT:  o->optn_ptr[0]=o->default_value; break;
        case OPTN_TYPE_TEXT: o->optn_ptr=NULL; break;
      }
  config_file=DEFAULT_CONFIG_FILE;
  sprite_origin="0,0";
  benchmark=0;
}

int preparse_cmdline(int argc, char *argv[]) {
  int i;
  for(i=1; i<argc; ++i) {
    if((i<argc-1) && (streq(argv[i],"-c") || streq(argv[i],"--config_file"))) {
      config_file=argv[i+1];
      return 0;
    }
    if(streq(argv[i],"-h") || streq(argv[i],"--help")) {
      show_help();
      return 1;
    }
  }
  return 0;
}


///////////////////////////////////////////////////////////////////////////////


static option_t* find_long_option(const char *where, const char *option) {
  option_t *o=options;
  for(o=options; o->optn_type; ++o)
    if(streq(o->long_name,option))
      return o;
  printf("%s: unknown option `%s'\n",where,option);
  return NULL;
}

static option_t* find_short_option(const char *where, char option) {
  option_t *o=options;
  for(o=options; o->optn_type; ++o)
    if(o->short_name==option)
      return o;
  printf("%s: unknown option `-%c'\n",where,option);
  return NULL;
}


static int set_option(const char *where, option_t* option, char *value, int is_volatile) {
  switch(option->optn_type) {
    case OPTN_TYPE_INT: {
      char *pos=value;
      int i;
      // check for valid number
      if(pos[0]=='-') ++pos;
      for(; *pos; ++pos)
        if(!isdigit(*pos)) {
          printf("%s: illegal value `%s' for option `%s': not a number\n",where,value,option->long_name);
          return 1;
        }
      // check if value is within bounds
      i=atoi(value);
      if(i<option->min) {
        printf("%s: illegal value `%s' for option `%s': too small\n",where,value,option->long_name);
        return 1;
      }
      if(i>option->max) {
        printf("%s: illegal value `%s' for option `%s': too large\n",where,value,option->long_name);
        return 1;
      }
      option->optn_ptr[0]=i;
      break;
    }
    case OPTN_TYPE_TEXT: {
      if(is_volatile) {
        char *safe=malloc(strlen(value)+1);
        if(safe) strcpy(safe,value);
        option->text_ptr[0]=safe;
      } else
        option->text_ptr[0]=value;
      break;
    }
  }
  return 0;
}


#define LINESIZE 80
void load_config(void) {
  FILE *cfg=fopen(config_file,"r");
  char line[LINESIZE];
  char *name,*value,*pos;
  char where[LINESIZE];
  int lineno=0;
  option_t *option;

  if(!cfg) {
    printf("load_config(): cannot open configuration file %s\n",config_file);
    return;
  }

  while(!feof(cfg)) {
    fgets(line,LINESIZE,cfg);
    if(feof(cfg)) break;
    line[LINESIZE-1]='\0';
    snprintf(where,LINESIZE,"load_config(): %s:%d",config_file,++lineno);
    where[LINESIZE-1]='\0';

    // trim away comments
    for(pos=line; *pos; ++pos)
      if(*pos=='#') {
        *pos='\0';
        break;
      }

    // extract name and value
    for(name=line; *name && isspace(*name); ++name);
    for(value=name; *value && !isspace(*value); ++value);
    for(; *value && isspace(*value); ++value) *value='\0';
    for(pos=value; *pos && !isspace(*pos); ++pos);
    *pos='\0';

    // check option name
    if(!name[0]) continue;
    option=find_long_option(where,name);
    if(!option) continue;

    // check option value
    if(!value[0]) {
      printf("%s: no value for option `%s'\n",where,name);
      continue;
    }
    set_option(where,option,value,1);
  }

  fclose(cfg);
}

int parse_cmdline(int argc, char *argv[]) {
  int argp;
  char *args;
  option_t *option;
  for(argp=1; argp<argc; ++argp) {
    args=argv[argp];

    // get suitable option
    if(args[0]=='-' && args[1] && !args[2])
      option=find_short_option("parse_cmdline()",args[1]);
    else if(args[0]=='-' && args[1]=='-' && args[2])
      option=find_long_option("parse_cmdline()",&args[2]);
    else {
      printf("parse_cmdline(): illegal command line option `%s'\n",args);
      return 1;
    }
    if(!option) return 1;

    // read value
    if(++argp>=argc) {
      printf("parse_cmdline(): no value for option `%s'\n",option->long_name);
      return 1;
    }
    if(set_option("parse_cmdline()",option,argv[argp],0))
      return 1;
  }
  return 0;
}
