/*
Sprite overlay module for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#include "options.h"
#include "sprite.h"

static pthread_mutex_t slock=PTHREAD_MUTEX_INITIALIZER;
static int sprite_valid=0;
static int sprite_start_x=0,sprite_start_y=0;
static int sprite_stride,sprite_size_x,sprite_size_y;
static uint8_t *sprite_data=NULL,*sprite_start=NULL;


int sprite_init(void) {
  char num1[10],*num2;  
  if(!sprite_file || !sprite_origin) return 0;

  // split sprite offset
  strncpy(num1,sprite_origin,10);
  num1[9]='\0';
  for(num2=num1; *num2 && *num2!=','; ++num2);
  if(!*num2) {
    printf("sprite_init(): invalid origin value `%s'\n",num1);
    return 1;
  }
  *num2++='\0';
  
  // get sprite offset values
  sprite_start_x=atoi(num1);
  sprite_start_y=atoi(num2);
  if(sprite_start_x<0 || sprite_start_x>=IN_WIDTH
  || sprite_start_y<0 || sprite_start_y>=IN_HEIGHT) {
    printf("sprite_init(): invalid origin value `%s,%s'\n",num1,num2);
    return 1;
  }

  // load sprite
  if(read_sprite()) {
    printf("sprite_init(): cannot read sprite file `%s' (bad format?)\n",sprite_file);
    return 1;
  }
  return 0;
}


int read_sprite(void) {
  FILE *f;
  uint16_t header[539];
  int i,size,oheight;

  if(!sprite_file) return 0;

  // lock sprite and invalidate
  pthread_mutex_lock(&slock);
  sprite_valid=0;
#define INVALID_SPRITE() \
  do { \
    if(f) fclose(f); \
    pthread_mutex_unlock(&slock); \
    return 1; \
  } while(0)

  // open file
  f=fopen(sprite_file,"rb");
  if(!f) INVALID_SPRITE();

  // read header
  i=fread(header,2,539,f);
  if(i!=539) INVALID_SPRITE();

  // parse header
  if(header[0]!=0x4D42  // "BM" magic
  || header[5]!=1078    // header size
  || header[13]!=1      // single plane
  || header[14]!=8)     // 8 bpp
    INVALID_SPRITE();
  sprite_size_x=header[9];
  sprite_size_y=oheight=header[11];
  sprite_stride=(sprite_size_x+3)&0x7FFFFFFC;

  // allocate and read image data
  size=sprite_stride*(sprite_size_y-1)+sprite_size_x;
  sprite_data=realloc(sprite_data,size);
  if(!sprite_data) INVALID_SPRITE();
  if(fread(sprite_data,1,size,f)!=size) INVALID_SPRITE();
  fclose(f);

  // adjust sprite size so that the image fits the screen
  i=IN_WIDTH -sprite_start_x; if(sprite_size_x>i) sprite_size_x=i;
  i=IN_HEIGHT-sprite_start_y; if(sprite_size_y>i) sprite_size_y=i;

  // flip the image (this is BMP after all ... :)
  sprite_start=&sprite_data[(oheight-1)*sprite_stride];
  sprite_stride=-sprite_stride;

  sprite_valid=1;
  pthread_mutex_unlock(&slock);
  return 0;
}



void draw_sprite(uint8_t *image, int stride) {
  if(!sprite_file) return;
  pthread_mutex_lock(&slock);
  if(sprite_valid) {
    int x,y;
    for(y=0; y<sprite_size_y; ++y) {
      uint8_t *sp=&sprite_start[y*sprite_stride];
      uint8_t *im=&image[(y+sprite_start_y)*stride+sprite_start_x];
#ifdef USE_X86_ASM
  asm("								\n\
    movl %2, %%ecx		# load counter and divide by 8	\n\
    shrl $3, %%ecx						\n\
    movl %0, %%esi		# load sprite address		\n\
    movl %1, %%edi		# load image address		\n\
    movl $8, %%eax		# load constant value 8		\n\
    .spriteloop:						\n\
    movq (%%edi), %%mm0		# load image value		\n\
    movq (%%esi), %%mm1		# load sprite value		\n\
    paddusb %%mm1, %%mm0	# do it				\n\
    movq %%mm0, (%%edi)		# save value			\n\
    addl %%eax, %%esi		# increment addresses		\n\
    addl %%eax, %%edi						\n\
    decl %%ecx			# decrease counter		\n\
    jnz .spriteloop		# loop over			\n\
    emms			# reset MMX state		\n\
  " :
  : "m" (sp),           // %0: sprite address
    "m" (im),           // %1: image address
    "m" (sprite_size_x) // %2: length in pixels
  : "eax", "ecx", "esi", "edi"
  );
#else
      for(x=0; x<sprite_size_x; ++x) {
        register int i=im[x]+sp[x];
        im[x]=(i&0x7FFFFF00)?255:i;
      }
#endif
    }
  }
  pthread_mutex_unlock(&slock);
}
