/*
Stream output module for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#include "options.h"
#include "stream.h"
#include "audio.h"
#include "ogmwrite.h"

int lo_current_au=0;
int hi_current_au=0;

static pthread_mutex_t streamlock=PTHREAD_MUTEX_INITIALIZER;

static ogg_phys_stream_t lo_phys,hi_phys;
static ogg_log_stream_t *lo_vid,*lo_aud,*hi_vid,*hi_aud;
static FILE *lo_file=NULL,*hi_file=NULL;
static FILE *lo_seg=NULL,*hi_seg=NULL;
static FILE *lo_idx=NULL,*hi_idx=NULL;

#define FILENAME_LENGTH 256
#define CURRENT_AU_FILE "current_au"
static char output_name[FILENAME_LENGTH+16];
static char *output_name_suffix;


static int create_dir(const char *dir) {
  if(!dir || !dir[0]) return 1;  // no or empty directory name? can't be real
  if(!mkdir(dir,0755)) return 0;  // try to create the directory
  if(errno==EEXIST) return 0;  // already exists? that's fine
  if(errno==ENOENT) {  // parent directory does not exist? so create it first
    int last_slash=strlen(dir);
    char *subdir;
    int res;
    // trim trailing slash first
    if(last_slash && dir[last_slash-1]=='/') --last_slash;
    // find last slash
    while(last_slash && dir[last_slash-1]!='/') --last_slash;
    if(!last_slash) return 1;  // already at toplevel? that's silly ...
    // copy truncated path
    subdir=alloca(last_slash);
    memcpy(subdir,dir,last_slash);
    subdir[last_slash-1]='\0';
    // create parent directory
    res=create_dir(subdir);
    if(res) return res;
    // finally, create the directory itself
    return mkdir(dir,0755);
  }
  return 1;  // every other error is critical
}



static void ogg_write_func(uint8_t *data, int length, void *priv) {
  if(!priv || !data || !length) return;
  if(priv==&lo_phys) {
    if(output_dir && lo_seg)
      fwrite(data,length,1,lo_seg);
    if(ogm_output && lo_file)
      fwrite(data,length,1,lo_file);
  }
  if(priv==&hi_phys) {
    if(output_dir && hi_seg)
      fwrite(data,length,1,hi_seg);
    if(ogm_output && hi_file)
      fwrite(data,length,1,hi_file);
  }
}


int stream_init(void) {
  // open segmented output files
  if(output_dir && *output_dir) {
    struct stat st;
    // copy output directory name
    strncpy(output_name,output_dir,FILENAME_LENGTH);
    output_name[FILENAME_LENGTH]='\0';
    output_name_suffix=&output_name[strlen(output_name)];
    // append slash if necessary
    if(output_name_suffix[-1]!='/')
      *output_name_suffix++='/';
    // check if there already is a stream
    strcpy(output_name_suffix,"hi/0");
    if(!stat(output_name,&st) || errno==EACCES) {
      printf("stream_init(): refusing to overwrite existing data in `%s'\n",output_dir);
      return 1;
    }
    // create directories
    umask(022);
    strcpy(output_name_suffix,"lo");
    if(create_dir(output_name)) {
      printf("stream_init(): cannot create output directory `%s'\n",output_name);
      return 1;
    }
    strcpy(output_name_suffix,"hi");
    if(create_dir(output_name)) {
      printf("stream_init(): cannot create output directory `%s'\n",output_name);
      return 1;
    }
    // create files -- step 1: lo-res current AU
    strcpy(output_name_suffix,"lo/" CURRENT_AU_FILE);
    lo_seg=fopen(output_name,"wb");
    if(!lo_seg) {
      printf("stream_init(): cannot create output file `%s'\n",output_name);
      return 1;
    }
    // create files -- step 2: lo-res index
    strcpy(output_name_suffix,"lo/index");
    lo_idx=fopen(output_name,"w");
    if(!lo_idx) {
      printf("stream_init(): cannot create index file `%s'\n",output_name);
      return 1;
    }    
    // create files -- step 3: hi-res current AU
    strcpy(output_name_suffix,"hi/" CURRENT_AU_FILE);
    hi_seg=fopen(output_name,"wb");
    if(!hi_seg) {
      printf("stream_init(): cannot create output file `%s'\n",output_name);
      return 1;
    }
    // create files -- step 4: hi-res index
    strcpy(output_name_suffix,"hi/index");
    hi_idx=fopen(output_name,"w");
    if(!hi_idx) {
      printf("stream_init(): cannot create index file `%s'\n",output_name);
      return 1;
    }    
  }

  // open plain output files
  if(ogm_output) {
    // prepare for appending suffixes
    char filename[FILENAME_LENGTH+8],*eon;
    strncpy(filename,ogm_output,FILENAME_LENGTH);
    filename[FILENAME_LENGTH]='\0';
    eon=&filename[strlen(filename)];
    // create output files
    strcpy(eon,"_lo.ogm");
    lo_file=fopen(filename,"wb");
    if(!lo_file)
      printf("stream_init(): warning: cannot write lo-res output file `%s'\n",filename);
    strcpy(eon,"_hi.ogm");
    hi_file=fopen(filename,"wb");
    if(!hi_file)
      printf("stream_init(): warning: cannot write hi-res output file `%s'\n",filename);
  }

  // initialize lo-res stream
  lo_phys.page_size=0;
  lo_phys.write_func=ogg_write_func;
  lo_phys.write_func_priv=(void*)(&lo_phys);
  ogg_phys_stream_init(&lo_phys);
  lo_vid=ogm_video_stream_create(&lo_phys,0,"XVID",FRAMERATE,1,LO_WIDTH,LO_HEIGHT,16000);
  lo_aud=ogm_audio_stream_create(&lo_phys,1,"0055",aud_samplerate,aud_channels,aud_bitrate*125,384,expected_frame_size);

  // initialize hi-res stream
  hi_phys.page_size=0;
  hi_phys.write_func=ogg_write_func;
  hi_phys.write_func_priv=(void*)(&hi_phys);
  ogg_phys_stream_init(&hi_phys);
  hi_vid=ogm_video_stream_create(&hi_phys,0,"XVID",FRAMERATE,1,HI_WIDTH,HI_HEIGHT,64000);
  hi_aud=ogm_audio_stream_create(&hi_phys,1,"0055",aud_samplerate,aud_channels,aud_bitrate*125,384,expected_frame_size);

  if(!ogm_output && !output_dir)
    printf("stream_init(): warning: no outputs selected\n");
  return 0;
}

int stream_close(void) {
  stream_next_au(LO_VIDEO_STREAM,0.0,1);
  stream_next_au(HI_VIDEO_STREAM,0.0,1);
  if(lo_file) fclose(lo_file);
  if(hi_file) fclose(hi_file);
  if(lo_idx) fclose(lo_idx);
  if(hi_idx) fclose(hi_idx);
  return 0;
}

void stream_write(int stream, uint8_t *packet, int length, uint64_t timestamp) {
  pthread_mutex_lock(&streamlock);
  switch(stream) {
    case AUDIO_STREAM:
      ogg_log_stream_write(lo_aud,packet,length,timestamp);
      ogg_log_stream_write(hi_aud,packet,length,timestamp);
      break;
    case LO_VIDEO_STREAM:
      ogg_log_stream_write(lo_vid,packet,length,timestamp);
      break;
    case HI_VIDEO_STREAM:
      ogg_log_stream_write(hi_vid,packet,length,timestamp);
      break;
  }
  pthread_mutex_unlock(&streamlock);
}

void stream_next_au(int stream, double timestamp, int eos) {
  ogg_phys_stream_t *ogg;
  int *current_au;

  pthread_mutex_lock(&streamlock);
  ogg=(stream==LO_VIDEO_STREAM)?(&lo_phys):(&hi_phys);
  current_au=(stream==LO_VIDEO_STREAM)?(&lo_current_au):(&hi_current_au);

  ogg_phys_stream_flush(ogg,eos);
  if(output_dir) {
    FILE **f=(stream==LO_VIDEO_STREAM)?(&lo_seg):(&hi_seg);
    // let's do some serious kick-ass pathname manipulations ;)
    char newname[FILENAME_LENGTH+16];
    sprintf(output_name_suffix,"%s/%d",(stream==LO_VIDEO_STREAM)?"lo":"hi",current_au[0]);
    strcpy(newname,output_name);
    strcpy(output_name_suffix,(stream==LO_VIDEO_STREAM)?("lo/" CURRENT_AU_FILE):("hi/" CURRENT_AU_FILE));
    // close current_au file, rename it, and re-open it
    if(*f) { fclose(*f); *f=NULL; }
    rename(output_name,newname);
    if(!eos)
      *f=fopen(output_name,"wb");
    // write index entry
    f=(stream==LO_VIDEO_STREAM)?(&lo_idx):(&hi_idx);
    if(*f) {
      if(eos) fprintf(*f,"%d\t\n",current_au[0]+1);
         else fprintf(*f,"%d\t%.3f\n",current_au[0]+1,timestamp);
      fflush(*f);
    }
  }

  current_au[0]++;
  pthread_mutex_unlock(&streamlock);
}
