/*
Video processing module for lvstream.
lvstream (C) 2005 Martin J. Fiedler <martin.fiedler@gmx.net>

lvstream is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

lvstream is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with lvstream; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include "main.h"
#include "options.h"
#include "capture.h"
#include "video.h"
#include "dvdecode.h"
#include "xvidwrap.h"
#include "stream.h"
#include "sprite.h"

static pthread_t thread=0;

static xvid_t lo_xvid,hi_xvid;
static int lo_next_au=0, hi_next_au=0;
static int next_keyframe=0;
static xvid_mv_t mvs[BLOCKWIDTH*BLOCKHEIGHT];

static int qadjust_next_frame=0;


///////////////////////////////////////////////////////////////////////////////


static void video_encode(xvid_t *xvid, int stream, int frameno, int *next_au) {
  uint8_t *data; int size;
  data=xvid_encode_frame(xvid,&size);
  if(!data) return;
  if(xvid->stats.type==XVID_TYPE_IVOP) {
    data[-1]=0x08;
    if(frameno>=*next_au) {
      stream_next_au(stream,frameno/(1.0*FRAMERATE),0);
      *next_au=frameno+au_length*FRAMERATE;
    }
  } else
    data[-1]=0;
  stream_write(stream,&data[-1],size+1,frameno);
}


///////////////////////////////////////////////////////////////////////////////


static void downscale(uint8_t *src, int src_stride,
                      uint8_t *dest, int dest_stride,
                      int dest_width, int dest_height
) {
  int x,y;
  uint8_t *ps,*pd;
  for(y=0; y<dest_height; ++y) {
    ps=&src[2*y*src_stride];
    pd=&dest[y*dest_stride];
    for(x=dest_width; x; --x) {
      register int i=(ps[0]+ps[1]+ps[src_stride]+ps[src_stride+1]+2)>>2;
      *pd++=i;
      ps+=2;
    }
  }
}


static void mv_upscale(xvid_mv_t *src, int src_stride,
                       xvid_mv_t *dest, int dest_stride,
                       int src_width, int src_height
) {
  int x,y;
  xvid_mv_t *ps,*pd;
  for(y=0; y<src_height; ++y) {
    ps=&src[y*src_stride];
    pd=&dest[2*y*dest_stride];
    for(x=src_width; x; --x) {
      xvid_mv_t mv=*ps++;
      mv.x*=2; mv.y*=2;
      pd[0]=mv;
      pd[1]=mv;
      pd[dest_stride]=mv;
      pd[dest_stride+1]=mv;
      pd+=2;
    }
  }
}


static void deinterlace(uint8_t *img, int stride, int width, int height) {
  int thresh=256-deint_strength;
  int x,y;
  for(y=1; y<height; y+=2) {
    uint8_t *line=&img[y*stride];
    for(x=0; x<width; ++x, ++line) {
      register int avg=(line[-stride]+line[stride])>>1;
      if(abs(avg-line[0])>=thresh) line[0]=avg;
    }
  }
}


///////////////////////////////////////////////////////////////////////////////


static int get_mvs(void * handle, int opt, void * param1, void * param2) {
  switch(opt) {
    case XVID_PLG_INFO: {
      xvid_plg_info_t *info=(xvid_plg_info_t*)param1;
      info->flags=XVID_REQ_MV_OUT;
      return 0;
    }
    case XVID_PLG_AFTER: {
      xvid_plg_data_t *data=(xvid_plg_data_t*)param1;
      if(data->mv_stride!=BLOCKWIDTH2 || data->mv_height!=BLOCKHEIGHT2)
        memset(mvs,0,sizeof(mvs));
      else
        mv_upscale(data->mvs,data->mv_stride, mvs,BLOCKWIDTH, BLOCKWIDTH2,BLOCKHEIGHT2);
      return 0;
    }
    case XVID_PLG_CREATE :
      *((void**)param2) = NULL;
    default:
      return 0;
  }
}


static int set_mvs(void * handle, int opt, void * param1, void * param2) {
  switch(opt) {
    case XVID_PLG_INFO: {
      xvid_plg_info_t *info=(xvid_plg_info_t*)param1;
      info->flags=XVID_REQ_MV_IN;
      return 0;
    }
    case XVID_PLG_FRAME: {
      xvid_plg_data_t *data=(xvid_plg_data_t*)param1;
      int dsize=data->mv_stride*data->mv_height*sizeof(xvid_mv_t);
      memcpy(data->mvs,mvs,dsize);
      return 0;
    }
    case XVID_PLG_CREATE :
      *((void**)param2) = NULL;
    default:
      return 0;
  }
}


///////////////////////////////////////////////////////////////////////////////


static void* video_thread(void *arg) {
  static uint8_t scaledY[LO_WIDTH*LO_HEIGHT];
  static uint8_t scaledU[LO_WIDTH*LO_HEIGHT/4];
  static uint8_t scaledV[LO_WIDTH*LO_HEIGHT/4];
  frame_data inframe;
  int expected_frameno=0,frameno;

  while(Running) {
    // wait for a frame to decode, and decode it
    dv_frame *dv=video_get_frame();
    if(!Running) return NULL;
    if(!dv) continue;
    frameno=dv->frameno;
    dvdecode_decode_video(dv->data,&inframe);
    video_frame_done();
    if(!inframe.Y) continue;

    // write skipped-frame markers
    if(frameno>expected_frameno) {
      int i;
      lo_xvid.frame.input.csp=XVID_CSP_SKIP;
      hi_xvid.frame.input.csp=XVID_CSP_SKIP;
      lo_xvid.frame.type=XVID_TYPE_PVOP;
      hi_xvid.frame.type=XVID_TYPE_PVOP;
      for(i=expected_frameno; i<frameno; ++i) {
        video_encode(&lo_xvid,LO_VIDEO_STREAM,i,&lo_next_au);
        video_encode(&hi_xvid,HI_VIDEO_STREAM,i,&hi_next_au);
      }
    }
    expected_frameno=frameno+1;

    // apply some filters
    if(deint_strength) {
      deinterlace(inframe.Y,inframe.Ystride,IN_WIDTH,  IN_HEIGHT);
      deinterlace(inframe.U,inframe.Cstride,IN_WIDTH/2,IN_HEIGHT/2);
      deinterlace(inframe.V,inframe.Cstride,IN_WIDTH/2,IN_HEIGHT/2);
    }
    draw_sprite(inframe.Y,inframe.Ystride);

    // downscale the raw images
    downscale(&inframe.Y[STARTX],  inframe.Ystride,scaledY,LO_WIDTH,  LO_WIDTH,  LO_HEIGHT);
    downscale(&inframe.U[STARTX/2],inframe.Cstride,scaledU,LO_WIDTH/2,LO_WIDTH/2,LO_HEIGHT/2);
    downscale(&inframe.V[STARTX/2],inframe.Cstride,scaledV,LO_WIDTH/2,LO_WIDTH/2,LO_HEIGHT/2);

/*
    // QUALITY TESTING CODE: write raw frames
{ FILE *f; char fn[256]; int y;
sprintf(fn,"/mnt/samt/500er/Incoming/lvstream-bench/lvstream_in_hi/%08d.pgmyuv",frameno+1);
f=fopen(fn,"wb");
if(!f) printf("fopen failed!\n");
fprintf(f,"P5 %d %d 255\n",HI_WIDTH,HI_HEIGHT*3/2);
for(y=0; y<HI_HEIGHT; ++y) fwrite(&inframe.Y[y*inframe.Ystride+STARTX],1,HI_WIDTH,f);
for(y=0; y<HI_HEIGHT/2; ++y) {
  fwrite(&inframe.U[y*inframe.Cstride+STARTX/2],1,HI_WIDTH/2,f);
  fwrite(&inframe.V[y*inframe.Cstride+STARTX/2],1,HI_WIDTH/2,f);
}
fclose(f);
sprintf(fn,"/mnt/samt/500er/Incoming/lvstream-bench/lvstream_in_lo/%08d.pgmyuv",frameno+1);
f=fopen(fn,"wb");
if(!f) printf("fopen failed!\n");
fprintf(f,"P5 %d %d 255\n",LO_WIDTH,LO_HEIGHT*3/2);
for(y=0; y<LO_HEIGHT; ++y) fwrite(&scaledY[y*LO_WIDTH],1,LO_WIDTH,f);
for(y=0; y<LO_HEIGHT/2; ++y) {
  fwrite(&scaledU[y*LO_WIDTH/2],1,LO_WIDTH/2,f);
  fwrite(&scaledV[y*LO_WIDTH/2],1,LO_WIDTH/2,f);
}
fclose(f);
}
*/

    // adjust quality once every second
    if(dv_frameno>=qadjust_next_frame) {
      qadjust_next_frame=dv_frameno+FRAMERATE;
      // if more framedrops occured than tolerated, decrease ME complexity
      if(drops_last_second>(FRAMERATE*framedrop_tolerance/100)
      && me_quality>0)
        video_set_quality(--me_quality);
    }

    // encode the frame in low resolution
    xvid_set_image(&lo_xvid,scaledY,scaledU,scaledV,LO_WIDTH,LO_WIDTH/2);
    if(force_keyframe_interval) {
      if(frameno>=next_keyframe) {
        lo_xvid.frame.type=XVID_TYPE_IVOP;
        next_keyframe=frameno+keyframe_interval;
      } else
        lo_xvid.frame.type=XVID_TYPE_PVOP;
    }
    video_encode(&lo_xvid,LO_VIDEO_STREAM,frameno,&lo_next_au);

    // encode the frame in high resolution
    xvid_set_image(&hi_xvid,&inframe.Y[STARTX],&inframe.U[STARTX/2],&inframe.V[STARTX/2],inframe.Ystride,inframe.Cstride);
    hi_xvid.frame.type=lo_xvid.stats.type;  // same frame type as in low res
    video_encode(&hi_xvid,HI_VIDEO_STREAM,frameno,&hi_next_au);
  }
  return NULL;
}


///////////////////////////////////////////////////////////////////////////////


void video_set_quality(int quality) {
  if(quality<1) {
    // -q 0: worst lo-res ME, no high-res refinement
    me_quality=0;
    xvid_set_me_preset(&lo_xvid,0);
    hi_xvid.frame.vop_flags=0;
    hi_xvid.frame.motion=XVID_ME_HINTED|XVID_ME_REFINE_ONLY;
  } else if(quality>8) {
    // -q 9: best lo-res ME, full (and best) hi-res ME
    me_quality=9;
    xvid_set_me_preset(&lo_xvid,7);
    xvid_set_me_preset(&hi_xvid,7);
  } else {
    // -q 1..8: variable lo-res ME, high-res scaling and refinement
    me_quality=quality;
    xvid_set_me_preset(&lo_xvid,quality-1);
    hi_xvid.frame.vop_flags=XVID_VOP_HALFPEL;
    hi_xvid.frame.motion=XVID_ME_HALFPELREFINE16|
                         XVID_ME_HINTED|XVID_ME_REFINE_ONLY;
  }
}


///////////////////////////////////////////////////////////////////////////////


int video_init(void) {
  char *ver=xvid_init();

  if(!ver) {
    printf("video_init(): cannot initialize XviD\n");
    return 1;
  }
  printf("video_init(): using %s\n",ver);

  printf("video_init(): creating lo-res XviD instance\n");
  xvid_prepare(&lo_xvid,LO_WIDTH,LO_HEIGHT,0,keyframe_interval,FRAMERATE,1);
  xvid_add_cbr_plugin(&lo_xvid,lo_bitrate);
  xvid_add_plugin(&lo_xvid,get_mvs,NULL,0);
  if(!xvid_create(&lo_xvid)) {
    printf("video_init(): cannot create lo-res XviD instance\n");
    return 1;
  }

  printf("video_init(): creating hi-res XviD instance\n");
  xvid_prepare(&hi_xvid,HI_WIDTH,HI_HEIGHT,0,keyframe_interval,FRAMERATE,1);
  xvid_add_cbr_plugin(&hi_xvid,hi_bitrate);
  xvid_add_plugin(&hi_xvid,set_mvs,NULL,0);
  if(!xvid_create(&hi_xvid)) {
    printf("video_init(): cannot create hi-res XviD instance\n");
    return 1;
  }

  video_set_quality(me_quality);

  // setup aspect ratios
#ifdef PAL
  lo_xvid.frame.par=XVID_PAR_43_PAL;
  hi_xvid.frame.par=XVID_PAR_43_PAL;
#endif
#ifdef NTSC
  lo_xvid.frame.par=XVID_PAR_43_NTSC;
  hi_xvid.frame.par=XVID_PAR_43_NTSC;
#endif

  return 0;
}

int video_start(void) {
  return pthread_create(&thread,NULL,video_thread,NULL);
}

int video_stop(void) {
  void *result;
  video_signal_exit();
  return pthread_join(thread,&result);
}

int video_close(void) {
  return 0;
}
